// --- [ rs232_2_twi.c ] ------------------------------------------------------
//       14.11.2008   Tab = 3
//
//		   Beschreibung ausgelagert in "rs232_2_twi.txt"
//
//       berarbeitung: 13.11.11
//       TWI-Slave Funktion ergnzt/korrigiert
//       Ausgabe der Frequenz bei -x korrigiert
//       berarbeitung: 21.11.11
//       - Abstrz beim Senden von Textdateien ist korrigiert.
//       - Fehler, wenn whrend der Zahleneingabe nach einer fhrenden '0' kein
//         weitere Ziffer oder ein x,b oder w folgte (es fehlte ein Break).
//       nderungen:
//       Empfang und Senden Serieller Daten erfolgt jeweils ber ISR.
//       Der Sendeteil erhlt einen eigenen Sendepuffer.
//       Beim Senden von Textdateien wird ein Echo automatisch unterdrckt, 
//       wenn noch eine bertragung aktiv ist,
//       es wird nur ein '+' als Platzhalter fr die fehlende Zeile ausgegeben.
//       Als Zeilenabschluss bei der Eingabe ist nun auch ein CR-LF mglich.
//       Die Puffergren der unterschiedlichen Zwischenpuffer wurden gendert.
// ----------------------------------------------------------------------------

#define EXTERN
#include <avr/io.h>
#include <stdlib.h>
#include <stdio.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <inttypes.h>
#include <string.h>
#include <ctype.h>

#include "global.h"
#include "TWI_Master_Slave.h"
#include "rs232_2_twi.h"
#include "rs232_int_m8.h"

// ----------------------------------------------------------------------------
void TWI_2_RS232(void);

// ----------------------------------------------------------------------------
int main(void)
{
uint8_t temp;
uint8_t i, error;
uint8_t chr0_flag;							// am Ende eines Textstrings muss ein CHR0 geschrieben werden

echo     = TRUE;								// Echo der Eingabe zum Terminal zurcksenden (TxD-Echo AN)
format   = 10;									// Parameter fr itoa(): 2 = binr, 10 = dezimal, 16 = hexadezimal
last_chr = 0;									// volatile: letztes empfangenes Zeichen - wartet auf RETURN
chr_in   = 0;									// volatile: Index fr vom Terminal eingelesene Zeichen

rs232_init();

MCUCR = 1<<SE;									// Sleep enablen (im Idle-Modus)
ACSR  = 1<<ACD;								// Analog comparator disablen

sei();

twi_sla_format = 'a';						// a oder h / Ausgabe ASCII oder Hexadezimal
TWI_Initialise(TWI_SLA_ADR);				// Slave mit Adresse 4 initialisieren

help();											// Hilfstext mit Befehlen ausgeben

while (1)
	{
	TWI_SLA_Start_Transceiver();			// den Slave (wieder) initialisieren
													// ACHTUNG: wenn vorher nicht mindestens ein Byte ber die serielle
													// Schnittstelle ausgegeben wurde, hngt das Programm in der folgenden Zeile fest !
													// aber help() hat ja schon eine Ausgabe gemacht.
	while(!(UCSRA &(1<<UDRE)));			// warten, bis eine RS232-Ausgabe beendet, damit nicht zu frh
													// in den sleep-modus gefallen wird

// Eingabeschleife ------------------------------------------------------------
// liest die Terminaleingabe ein bis zu einem RET oder zum Ende des Puffers
// falls ein TWI-Interrupt sich ereignete, dann sind Daten via TWI-empfangen worden,
// die werden dann an TX bergeben
	do	{											// Sleep-Modus starten - sofern nicht zwischenzeitlich
													// per TWI Daten an den Slave gesendet wurden
		if (!(TWI_SLA_RxCount()))
			{										// warten, bis sich ein Interrupt ereignet
			asm volatile("sleep");			// RXC oder TWINT beendet das Warten
			}

		if (TWI_SLA_RxCount())				// wenn der TWI-Slave Daten vollstndig empfangen hat
			{
			TWI_2_RS232();						// aufbereiten und ausgeben
			TWI_SLA_Start_Transceiver();	// den Slave wieder initialisieren
			while(!(UCSRA &(1<<UDRE)));	// warten, bis die RS232-Ausgabe beendet, damit nicht zu frh
			}										// in den sleep-modus gefallen wird
								
		if (last_chr == 13)					// einlesen, bis RETURN oder Ende des Zeichenpuffers erreicht					
			{
			last_chr = 0;						// damit keine Endlosschleife entsteht ...
			break;								// do {} while verlassen
			}
		} while(chr_in < RS232_BUFFER);

// -> Die Eingabe ist mit einem RETURN abgeschlossen (oder der Puffer ist voll)

// Auswertung der Eingabe -----------------------------------------------------
// wandelt Ziffernfolgen in rs232_rx in Bytes um,  wirft die Trennzeichen "," raus und schreibt das
// Ergebnis in write_buf
// Kommandos (die mit "-" beginnen) werden separat behandelt.

	chr_out 	 = 0;								// Index fr aus dem Puffer ausgelesene Zeichen.
	chr_twi 	 = 0;								// Index fr auf den TWI-Bus auszugebende/lesende Zeichen
  	chr0_flag = 0;

	while((!(rs232_rx[chr_out] == 13)) && (chr_out <= chr_in))
		{
		if ((chr_out == 0) && (rs232_rx[0] == '-'))	// wenn das erste Zeichen = '-', dann folgen Kommandos
			{
			rs232_rx[1] = tolower(rs232_rx[1]);		// Parameter klein schreiben, danach nur auf Kleinschreibung prfen
			if (rs232_rx[1] == 'l')						// Parameter prfen
				{
				twi_test();
				}

			else if (rs232_rx[1] == 'x')				// TWI-Takt auf 100KHz
				{
  				TWBR = (F_CPU / (2 * 100000) - 8);
				show_twi_takt(F_CPU/(16 + 2 * (F_CPU / (2 * 100000) - 8)) / 1000);
				}

			else if (rs232_rx[1] == '0')	 			// TWBR einstellen
				{
  				TWBR = 10;									// TWBR should be 10 or higher
				show_twi_takt(F_CPU/(16 + 2 * 10) / 1000);
				}

			else if (rs232_rx[1] == '1')
				{
  				TWBR = 20;
				show_twi_takt(F_CPU/(16 + 2 * 20) / 1000);
				}

			else if (rs232_rx[1] == '2')
				{
				TWBR = (40);
				show_twi_takt(F_CPU/(16 + 2 * 40) / 1000);
				}

			else if (rs232_rx[1] == '3')
				{
				TWBR = 64;
				show_twi_takt(F_CPU/(16 + 2 * 64) / 1000);
				}

			else if (rs232_rx[1] == '4')
				{
				TWBR = 128;
				show_twi_takt(F_CPU/(16 + 2 * 128) / 1000);
				}

			else if (rs232_rx[1] == '5')
				{
				TWBR = 255;
				show_twi_takt(F_CPU/(16 + 2 * 255) / 1000);
				}

			else if (rs232_rx[1] == 'b')	// Ausgabe binaer
				{
				format = 2;
				show_format();
				}

			else if (rs232_rx[1] == 'd')	// Ausgabe dezimal
				{
				format = 10;
				show_format();
				}

			else if (rs232_rx[1] == 'h')	// Ausgabe hexadezimal
				{
				format = 16;
				show_format();
				}

			else if (rs232_rx[1] == '?')	// Hilfe ausgeben
				{
				help();
				}

			else if (rs232_rx[1] == 's')	// als TWI_Slave empfangene Daten hexadezimal/ascii ausgeben
				{
				puts_p(PSTR("Ausgabeformat TWI-Slave-Daten -> "));
				if (twi_sla_format == 'a') 
					{
					twi_sla_format++;			// a incrementieren zu b (!= a)
					puts_p(PSTR("NUMERISCH"));
					}
				else 
					{
					twi_sla_format = 'a';
					puts_p(PSTR("ASCII"));
					}
				put_cr(); 
				}
				
			else if (rs232_rx[1] == 'e')	// Eingabemodus / Echo togglen
				{
				echo ^= 0x01;
				show_echo();					// aktuelle Einstellung ausgeben
				}

			else									// alle anderen Parameter ignorieren
				{
				puts_p(PSTR("unbekannt: -"));// Fehlermeldung ausgeben
				put_c(rs232_rx[1]);
				put_cr();
				}

			chr_out = 0;						// der Auswertung mitteilen, dass keine Ausgabe vorliegt
			break;								// 
			}

		else if (rs232_rx[chr_out] == '+')	// ein '+' behlt den letzten Eintrag im Sendepuffer bei,
			{										// so muss nicht immer Adresse/Kommando neu eingegeben werden.
			chr_out++;							// Index Lesepuffer
			chr_twi++;							// Zhler fr die Anzahl der auszugebenden Zeichen
			}

		else if (isdigit(rs232_rx[chr_out])) // wenn eine Ziffer vorliegt, dann in ein Byte umwandeln lassen
			{										// Workaround fr 16-Bit Werte:
			temp = Zahl_read();				// Zahl_read() schreibt bei 16-Bit-Werten das High-Byte direkt in den write_buf[],
													// incrementiert den Index und liefert das Low-Byte zurck.
			write_buf[chr_twi] = temp;		// Die Schreibweise wie unten funktioniert daher nicht bei 16-Bit Werten !
			chr_twi++;							// dabei geht das High-Byte verloren !!
			}										// write_buf[chr_twi++] = Zahl_read(); -> funktioniert nicht !!!!

		else if(rs232_rx[chr_out] == ',')// ein Komma einfach bergehen ...
			{
			chr_out++;							// Index Lesepuffer incrementieren, Schreibpuffer nicht ndern
			}

		else										// wenn das nchste Zeichen keine Ziffer ist,
			{										// dann wird der Rest als Text interpretiert (bis zum RETURN)
			while((!(rs232_rx[chr_out] == 13)) && (chr_out <= chr_in))
				{
				if (rs232_rx[chr_out] == 92) // wenn ein Backslash im Puffer steht
					{
			      if (chr0_flag)          // wenn das Flag bereits gesetzt, dann einen CHR(0) schreiben
			      	{
			      	write_buf[chr_twi] = 0; // ein CHR(0) schreiben als Abschluss eines Strings
			      	chr_twi++;
			      	chr_out++;           // jetzt sollte eigentlich ein RETURN folgen - wird hier nicht geprft

			      	goto exit_read;	   // das weitere Einlesen wird abgebrochen, eine '0' innerhalb eines Textes macht keinen Sinn
			      	}

			      else							// wenn das Flag noch nicht gesetzt (am Anfang des Textblockes)
			      	{
			      	chr0_flag++;			// Flag setzen
			      	chr_out++;				// Zhler incrementieren, aber ohne eine Ausgabe in den Sendepuffer

			      								// damit ein einzelner Backslash am Ende einer Zeile ausreicht, um ein '0' zu senden
													// prfen, ob das Ende der Eingabe erreicht ist
			      	if (chr_out == (chr_in - 1))
			      		{
			      		write_buf[chr_twi++] = 0;
			      		goto exit_read;	// Einlesen beenden
			      		}
			     		}
					}

				else            				// wenn kein Backslash im Puffer, dann Zeichen einfach im Sendepuffer ablegen
					{
					write_buf[chr_twi++] = rs232_rx[chr_out++];
					}
		  		}
			}
		}

	exit_read:									// nicht 100%ig elegant ....

// Ausgabe auf den TWI- Bus ---------------------------------------------------

													// fr den Slave-Modus war TWIE gesetzt.
													// die Master-Routinen erkennen aber am gesetzten TWIE-Flag, dass noch
													// eine bertragung aktiv ist - und warten bis TWIE gelscht ist.
	TWCR &= ~(1<<TWIE);						// Um eine Endlosschleife zu vermeiden, muss der Slave-Modus beendet werden.
													// Der TWI-Master denkt sonst, dass Modul sei noch aktiv und wartet auf
													// das Zurcknehmen des Flags ...
													// Am Ende des Programmteils muss der Slave neu aktiviert werden !
	
	chr_in = 0; 								// den Index fr den Empfangspuffer der seriellen Schnittstelle zurckstellen
	
	if (chr_out)								// nur ausfhren, wenn > 0 (wenn berhaupt Eingaben vorhanden)
		{

// --- Daten lesen ------------------------------------------------------------
		if (write_buf[0] & 0x01)			// testen, ob Adressbit.0 gesetzt, dann wird vom TWI-Bus gelesen
			{
			if (chr_twi == 2)					// wenn genau 2 Byte eingelesen, dann sind das die Adresse + Anzahl der zu lesenden Bytes
				{
				chr_twi = write_buf[1] + 1;// fr die Adresse um 1 erhhen
													// aber die Arraygrenzen beachten !!
				if (chr_twi > TWI_BUFFER) chr_twi = TWI_BUFFER;
				}

			else if (chr_twi < 2)   		// sonst wird die Anzahl der Parameter gewhlt
				{
				chr_twi = 2;
				}

			else
				{
				chr_twi++;						// die Anzahl der Bytes + 1
				}
													// um die Eingabe nicht zu berschreiben, wird zum Empfang ein anderer Buffer
													// verwendet. Dadurch kann mit +++ die letzte Eingabe bernommen werden.
			read_buf[0] = write_buf[0];	// write_buf nach read_buf kopieren
		
			TWI_MA_Start_Transceiver_With_Data(read_buf, chr_twi);
			error = TWI_Get_Err();
													// die Antwort abholen, dazu das Ende der bertragung abwarten
			TWI_MA_Get_Data_From_Transceiver(read_buf, chr_twi);

			if (!(error))
				{
				puts_p(PSTR("RxD "));
				put_string(itoa(read_buf[0], msg, format));
				puts_p(PSTR(": "));
				for(i = 1; i< chr_twi; i++)
					{
					put_string(itoa(read_buf[i], msg, format));
					put_c(',');
					}
					put_cr();
				}

			else  									// wenn keine Daten empfangen wurden
				{
				puts_p(PSTR("Adr "));			// Gerteadresse ausgeben
				put_string(itoa(read_buf[0], msg, format));
				err_msg(error);					// Fehlercode/meldung ausgeben incl. Return
				}
			}

// --- Daten schreiben ---------------------------------------------------------
		else											// Adressbit.0 war 0, es wird geschrieben
			{
			TWI_MA_Start_Transceiver_With_Data(write_buf, chr_twi);
	
			error = TWI_Get_Err();				// Warteschleife, bis die bertragung beendet ist
														// vor dem Aktivieren des Slave-Modus einlesen
			
			if ((echo) && (UCSRB == TX_DISABLE))// nur dann ein Echo ausgeben, wenn kein Sendung aktiv
				{										// beim Senden einer Textdatei werden hier Teile
														// des Echos unterdrckt - es gibt aber keinen Absturz

				if (!(error))						// kein Fehler
					{
					puts_p(PSTR("TxD "));		// Gerteadresse ausgeben
					put_string(itoa(write_buf[0], msg, format));
					puts_p(PSTR(": "));
					for(i = 1; i < chr_twi; i++)
						{
						put_string(itoa(write_buf[i], msg, format));
						put_c(',');
						}
					put_cr();
					}

				else									// Fehlercode und Gerteadresse ausgeben
					{
					puts_p(PSTR("Adr "));
					put_string(itoa(write_buf[0], msg, format));
					err_msg(error);				// Fehlercode/meldung ausgeben incl. Return
					}
				}
				
			else if (UCSRB == TX_ENABLE)		// wenn eine Zeile unterdrckt wird, dann zumindest ein + ausgeben
				{
				put_c('+');
				put_c(13);
				}
			}
		}
	}
}

// ----------------------------------------------------------------------------
// zeigt den status von no_echo aus
// (Echo der Eingabe ber COM zurcksenden ans Terminal)
void show_echo(void)
{
puts_p(PSTR("Eingabe-Echo: "));					// und Status ausgeben
if (echo)
	{
	puts_p(PSTR("EIN"));
	}
else
	{
	puts_p(PSTR("AUS"));
	}
put_cr();
}

// ----------------------------------------------------------------------------
// gibt das Ausgabeformat aus, 'format' ist wird als Parameter in 'itoa()' benutzt
void show_format(void)
{
puts_p(PSTR("Ausgabeformat -> "));
if (format == 2)
	{
	puts_p(PSTR("binaer"));
	}
else if (format == 10)
	{
	puts_p(PSTR("dezimal"));
	}
else if (format == 16)
	{
	puts_p(PSTR("hexadezimal"));
	}
put_cr();
}

// ----------------------------------------------------------------------------
// gibt den TWI-Takt und TWBR aus
void show_twi_takt(uint8_t takt)
{
puts_p(PSTR("TWI-Takt = "));
put_string(itoa(takt, msg, 10));
puts_p(PSTR("KHz, TWBR = "));
put_string(itoa(TWBR, msg, 10));
put_cr();
}


// ----------------------------------------------------------------------------
// Die Meldung, die bei Einschalten ausgeben wird (und bei -?)
void help(void)
{
put_cr();
put_cr();
puts_p(PSTR("TERMINAL_2_TWI  -  Slave_Adr: 0x"));
put_string(utoa(TWI_SLA_ADR, msg, 16));
puts_p(PSTR("  -  21.11.11:"));
put_cr();
put_cr();
puts_p(PSTR("Eingabeformat:"));
put_cr();
puts_p(PSTR("255(dez.Byte), 0xFF(hex.Byte), 0b11111111(bin.Byte), 0w65536(dez.Word)"));
put_cr();
put_cr();
puts_p(PSTR("Befehle:"));
put_cr();
puts_p(PSTR("-[0|1|2|3|4|5] -> TWBR einstellen auf: [10|20|40|64|128|255]"));
put_cr();
puts_p(PSTR("-x      -> TWI-Takt auf 100KHz einstellen [= default]"));
put_cr();
puts_p(PSTR("-[b|d|h]-> Ausgabe [binaer|dezimal|hexdezimal]: "));
show_format();
puts_p(PSTR("-s      -> Ausgabe TWI-Slave-Daten [ascii|num]: "));
if (twi_sla_format == 'a') puts_p(PSTR("ASCII")); else puts_p(PSTR("NUM"));
put_cr();
puts_p(PSTR("-e      -> toggle "));				// den aktuellen Status schreibt show_echo()
show_echo();
puts_p(PSTR("           Fuer die Eingabe aus Text-Dateien AUS schalten !!"));
put_cr();
puts_p(PSTR("-l      -> Suche nach angeschlossenen TWI-Slaves"));
put_cr();
puts_p(PSTR("-?      -> diese Hilfe"));
put_cr();
put_cr();
}

// ----------------------------------------------------------------------------
// gibt Fehlercodes und Fehlermeldungen des TWI-Master aus
void err_msg(uint8_t error)
{
puts_p(PSTR(" TWI-Error 0x"));				//
put_string(itoa(error, msg, 16));			// Fehlercode immer hexadezimal
puts_p(PSTR(": "));								// Trennzeichen

if (error ==  0x08)
	{
	puts_p(PSTR("START has been transmitted"));
	}
else if (error ==  0x10)
	{
	puts_p(PSTR("Repeated START has been transmitted"));
	}
else if (error == 0x38)
	{
	puts_p(PSTR("Arbitration lost"));
	}
else if (error == 0x18 )
	{
	puts_p(PSTR("SLA+W has been transmitted and ACK received"));
	}
else if (error == 0x20 )
	{
	puts_p(PSTR("SLA+W has been transmitted and NACK received"));
	}
else if (error == 0x28 )
	{
	puts_p(PSTR("Data byte has been transmitted and ACK received"));
	}
else if (error ==  0x30)
	{
	puts_p(PSTR("Data byte has been transmitted and NACK received"));
	}
else if (error ==  0x40 )
	{
	puts_p(PSTR("SLA+R has been transmitted and ACK received"));
	}
else if (error == 0x48 )
	{
	puts_p(PSTR("SLA+R has been transmitted and NACK received"));
	}
else if (error == 0x50 )
	{
	puts_p(PSTR("Data byte has been received and ACK transmitted"));
	}
else if (error ==  0x58)
	{
	puts_p(PSTR("Data byte has been received and NACK transmitted"));
	}

//else if (error ==  0xF8)	// ist das Ergebnis eines erfolgreichen Transfers
//	{
//	puts_p(PSTR("No relevant state information available; TWINT = 0"));
//	}

else if (error ==  0x00)
	{
	puts_p(PSTR("Bus error due to an illegal START or STOP condition"));
	}
else
	{
	puts_p(PSTR("unknown error"));
	}

put_cr();
}
// ----------------------------------------------------------------------------
// Zeigt eine Liste aller am TWI-Bus antwortenden Gerte an
// wird durch die Eingabe von '-l' aufgerufen
void twi_test(void)
{
uint8_t i = 0;
uint8_t found = 0;

put_cr();
puts_p(PSTR("Suche nach TWI-Device ..."));
put_cr();
																// hier wird TWIE gelscht
TWCR &= ~(1<<TWIE);										// TWIE ist gesetzt worden fr SLA_read
																// TWI_MA_Start hngt dann in Endlosschleife
for (i=0; i < 128; i++)									// Adressen von 0 ... 254 Step 2
	{
	write_buf[0] = i<<1;									// Adresse in den Sendepuffer schreiben
																// dabei um eine Stelle nach links schieben

	TWI_MA_Start_Transceiver_With_Data(write_buf, 1 );	// ein Byte (die Adresse) senden
	_delay_ms(1);
																// Ergebnis testen auf TWI_NO_STATE
	if (!(TWI_Get_Err()))								// dann liegt kein Fehler vor
		{
		found++;												// Zhler fr gefundene Gerte
		if (format == 2) puts_p(PSTR("0b"));
		if (format == 16) puts_p(PSTR("0x"));
		put_string(itoa(write_buf[0], msg, format));// Adresse ausgeben
		put_c(',');											// Komma + Leerzeichen ausgeben
		put_c(' ');
		}
	}

if (found)													// Zusammenfassung am Ende
	{
	put_cr();												// neue Zeile
	puts_p(PSTR("gefunden: "));
	put_string(itoa(found, msg, 10));				// Print Anzahl Gerte gefunden
	}

else
	{
	put_cr();												// neue Zeile
	puts_p(PSTR("kein TWI-Device gefunden!"));	// Print keine Gerte gefunden
	}

put_cr();													// neue Zeile
}

// ----------------------------------------------------------------------------
// liest eine Ziffernfolge aus dem Puffer und gibt sie als ein Byte zurck
// die aufrufende Routine hat auf isdigit() geprft.
// die Ziffernfolge kann mit '0b', '0x' oder '0w' beginnen, dann wird in Unterprogramme
// zur Erkennung verzweigt.
// Achtung: die Rckgabe von 16Bit-Werten ist SEHR unkonventionell !
//          1 Byte wird zurckgegeben, 1Byte direkt in den Ausgangpuffer geschrieben

uint8_t Zahl_read(void)
{
uint8_t Zahl = 0;
uint8_t i = 0;
uint8_t zero = 0;											// wird TRUE, wenn die erste Ziffer eine 0 ist
																// fr die Eingabe im Format 0x../ 0b..
if (rs232_rx[chr_out] == '0') zero = TRUE;
do
	{
	if (isdigit(rs232_rx[chr_out]))					// solange einlesen, wie Ziffern vorhanden
		{
		Zahl *= 10;
		Zahl += (rs232_rx[chr_out++] - 0x30);
		i++;
		}
	else														// taucht ein Buchstabe auf, dann ...
		{
		if ((zero) && (i == 1))							// wenn das erste Zeichen eine '0' war und nun
			{													// das nchste (das zweite) Zeichen ansteht
			rs232_rx[chr_out] = toupper(rs232_rx[chr_out]); // in Groschreibung umwandeln
			if (rs232_rx[chr_out] == 'X')
				{
				return(hex_read());						// eine Hexadezimalzahl einlesen und zurckgeben
				}

			else if (rs232_rx[chr_out] == 'B')
				{
				return(bin_read());						// eine Binrzahl einlesen und zurckgeben
				}
																// Word 0w...
			else if (rs232_rx[chr_out] == 'W')
				{
				return(long_read());
				}
			else break;										// Abbruch auch dann, wenn nach der 0 ein Buchstabe auftritt
			}

		else													// ein Buchstabe taucht in der Ziffernfolge auf
			{
			break;											// abbrechen
			}
		}
	}
	while ((!(rs232_rx[chr_out] == ',')) && (!(rs232_rx[chr_out] == 13)) && (i < 3));
	// solange einlesen bis:
	// ',' oder RETURN eingelesen oder
	// mehr als drei Zeichen eingelesen (255!), berlauf wird ignoriert (256 wird zu 1)!!

	return(Zahl);
}

// ----------------------------------------------------------------------------
// liest eine Zahl vom Typ long/word mit bis zu 5 Stellen ein
// das high-byte schreibt die Funktion in den TWI-Buffer, das low-Byte wird zurckgegeben
uint8_t long_read(void)
{
uint16_t Zahl = 0;
uint8_t i = 0;

chr_out++;													// von 0l.. zur ersten Ziffer gehen
do
	{
	if (isdigit(rs232_rx[chr_out]))					// solange einlesen, wie Ziffern vorhanden
		{
		Zahl *= 10;
		Zahl += (rs232_rx[chr_out++] - 0x30);
		i++;
		}
	else
		{
		break;
		}
	}
while ((!(rs232_rx[chr_out] == ',')) && (!(rs232_rx[chr_out] == 13)) && (i < 5));

write_buf[chr_twi] = Zahl >> 8;						// schreibt hier das high-Byte,
chr_twi++;

return(Zahl);												// das low-Byte wird an die aufrufende Funktion zurckgegeben
}

// ----------------------------------------------------------------------------
// liest die nachfolgenden (bis zu) 8 Stellen als Binrzahl ein
// ein Zeichen <>[0|1] beendet die Funktion
// der Index chr_out steht noch auf "B" von 0b...
uint8_t bin_read(void)
{
uint8_t bin = 0;
uint8_t i = 0;

for (i = 0; i < 8; i++)
	{
	chr_out++;												// Index incrementieren
	if (rs232_rx[chr_out] == '0')						// bei einer Null nur nach links schieben
		{
		bin *= 2;											// um eine Stelle nach links schieben
		}

	else if (rs232_rx[chr_out] == '1')				// bei einer Eins nach links schieben
		{														// und 1 addieren
		bin *= 2;											// um eine Stelle nach links schieben
		bin++;												// und eine 1 addieren
		}

	else														// wenn weder 0 noch 1, dann
		{
		return(bin);										// den bisher ermittelten Wert zurckliefern,
		}														// den Index chr_out nicht verndern
	}

chr_out++;													// den Index auf das nchste Zeichen setzen
return(bin);
}

// ----------------------------------------------------------------------------
// liest zwei folgende Zeichen als Hexadezimalzahl ein
// ein Zeichen, das weder eine Ziffer noch [A-E] ist, beendet die Eingaben
// der Index chr_out steht noch auf dem "X" von 0x..
uint8_t hex_read(void)
{
uint8_t hex = 0;
uint8_t i = 0;
int8_t temp;

for (i = 0; i < 2; i++)
	{
	chr_out++;												// Index von rs232_rx auf nchstes Zeichen
	temp = toupper(rs232_rx[chr_out]);				// in Groschreibung umwandeln und puffern
																// rs232_rx knnte ja einen nachfolgenden Text enthalten ...
//	if ((temp > 47) && (temp < 58))					// auf Ziffer prfen
	if (isdigit(temp))
		{
		hex *= 16;
		hex += (temp - '0');								// Wenn Ziffer, dann zu Zahl machen
		}

	else if ((temp > ('A'- 1)) && (temp < ('F'+ 1)))	// Wenn 'A .. F'
		{
		hex *= 16;
		hex += (temp - 'A' + 10);						// in Ziffer umwandeln und addieren
		}

	else
		{
		return(hex);										// den bisher ermittelten Wert zurckliefern,
		}														// den Index chr_out nicht verndern
	}

chr_out++;													// den Index auf das nchste Zeichen setzen
return(hex);												// Wert zurckliefern
}


// ----------------------------------------------------------------------------
// sendet ber TWI an den Slave versandte Bytes direkt ber RS232 zum Terminal
// die Daten werden abgeholt, in Zeichenfolge umgewandelt und versandt
// das 1. Byte (= TWI-Adressse) wird weggelassen
// die Daten werden alternativ als Hex-Werte mit nachfolgendem Trennzeichen
// oder als ASCII-Zeichen ausgegeben (nicht druckbare Zeichen werden zu ' ' umgewandelt.

void TWI_2_RS232(void)
{
// TWI_SLA-Buffer einlesen, Bytes in Strings umwandeln und Ausgabestring generieren
char delimiter[2] = ",";

uint8_t i, j, x = 0;
rs232_tx[0] = 0;											// globale Variable, String lschen

j = TWI_SLA_RxCount();									// Anzahl der eingegangen Bytes abholen
TWI_SLA_Get_Data_From_Transceiver(read_buf, j);

for(i = 1; i < j; i++)									// das 1.Byte (TWI-Adresse) auslassen
	{
	if (twi_sla_format == 'a')
	{
		// Nicht druckbare Zeichen als Space darstellen, ein RETURN gibt keine neue Zeile aus!
		// ggf. muss auf read_buf[i] = 13 geprft und ein put_cr() gesendet werden.
		if ((read_buf[i] < 32) || (read_buf[i] > 127)) read_buf[i] = ' ';
		rs232_tx[x++] = read_buf[i];
	}
			
	else
	{
		strcat( rs232_tx, utoa(read_buf[i], msg, format));	// umwandeln in Format der Wahl
		strcat( rs232_tx, delimiter);					// und ein Trennzeichen anhngen
	}
}
// Bei der ASCII-Ausgabe den String noch terminieren ...
if (twi_sla_format == 'a') rs232_tx[x] = 0;

put_string(rs232_tx);
put_cr();

}

// --- [ eof ] ----------------------------------------------------------------