// --- [ rs232_int_m8.c ] -----------------------------------------------------
//
//       21.11.11 - fr rs232_2_twi mit ATMega8

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include "global.h"
#include "rs232_int_m8.h"

volatile uint8_t rs_buffer[TX_BUFFER_SIZE];	// Ringpuffer zum Senden
volatile uint8_t rs_tail = 0, rs_head = 0;

// ----------------------------------------------------------------------------
//
void rs232_init(void)
{
UBRRL = TEILER;									// siehe rs232.h
UBRRH = 0;
UCSRB  = TX_DISABLE;
UCSRC = (1<<URSEL | 0<<UMSEL) | (1<<UCSZ1) | (1<<UCSZ0);
}

// ----------------------------------------------------------------------------
// ein einzelnes Zeichen auf COM ausgeben
void put_c(uint8_t c)
{

register uint8_t x;
x = rs_head;

// verhindern, dass der Puffer von hinten her berschrieben wird
// wartet, bis rs_tail durch Senden incrementiert wurde
while (((x + 1) & TX_BUFFER_MASK) == rs_tail);

rs_buffer[x++] = c;								// c in die Warteschlange einreihen
x &= TX_BUFFER_MASK;								// Pointer-berlauf maskieren
rs_head = x;										// Register zurckschreiben

if (UCSRB == TX_DISABLE)						// Prfen, ob TX disabled ist
	{
	x = rs_tail;									// Tail einlesen
	UCSRB = TX_ENABLE;							// TX enablen

	UDR = rs_buffer[x++];						// und ein Byte auf die Reise schicken
	x &= TX_BUFFER_MASK;							// Pointer incrementieren, maskieren
	rs_tail = x;									// und speichern
	}
}

/*
// ----------------------------------------------------------------------------
// gibt ein einzelnes RETURN auf COM aus - ohne Pufferung und Interrupt
void put_c(uint8_t c)
{
while(!(UCSRA &(1<<UDRE)));						// warten bis sendebereit
UDR = c;
}
*/

// ----------------------------------------------------------------------------
// gibt ein einzelnes RETURN auf COM aus
void put_cr(void)
{
//while(!(UCSRA &(1<<UDRE)));						// warten bis sendebereit
//UDR = 13;
put_c(13);
}

// ----------------------------------------------------------------------------
// einen String auf COM ausgeben
void put_string(char *s)
{
while(*s) put_c(*s++);
}

// ----------------------------------------------------------------------------
// gibt einen String aus dem Flash aus
void puts_p(const char *progmem_s)
{
register char c;
while ( (c = pgm_read_byte(progmem_s++)) )
	{
   put_c(c);
	}
}

// ----------------------------------------------------------------------------
// Nachdem ein Byte verschickt worden ist, wird der TXC-Interrufpt aufgerufen.
// Es wird geprft, ob weitere Daten im Ringpuffer sind.
// Wenn ja, dann wird das nchste Byte ausgegeben,
// wenn nicht, dann wird die bertragung beendet.
// Der Status von UCSRB gibt zu erkennen, ob eine bertragung aktiv ist.
ISR(USART_TXC_vect)
{

register uint8_t r;								// volatile rs_tail in ein Register laden
r = rs_tail;

if (!(r == rs_head))								// wenn rs_tail == rs_head, dann ist der Puffer leer
	{
	UDR = rs_buffer[r++];						// Byte in den Sender schreiben
	r &= TX_BUFFER_MASK;							// Pointer incrementieren und maskieren
	rs_tail = r;									// und zurckschreiben
	}

else													// der Puffer ist leer, bertragung beenden
	{
	UCSRB = TX_DISABLE;							// TXC-Interrupt disablen
	}
}

// ----------------------------------------------------------------------------
ISR(USART_RXC_vect)
{
uint8_t temp;
			
	temp = UDR;									// ohne den Umweg ber temp wird UDR nicht gelesen ???
	if ((temp > 29) || (temp == 13))		// alle Steuerzeichen ausser RETURN ignorieren
		{
		rs232_rx[chr_in++] = temp;			// ASCII-Wert einlesen, im Puffer ablegen, Index incr
		}
	last_chr = temp;							// damit in main() auf RETURN geprft werden kann
}

// --- [ eof ] ----------------------------------------------------------------
